from importlib.metadata import version
from typing import Union

from e2b.constants import API_HOST
from e2b.api.metadata import default_headers


pydantic_version = version("pydantic")
if pydantic_version < "2.0.0":
    import e2b.api.v1.client as client
    import e2b.api.v1.client.models as models
    import e2b.api.v1.client.exceptions as exceptions
else:
    import e2b.api.v2.client as client
    import e2b.api.v2.client.models as models
    import e2b.api.v2.client.exceptions as exceptions


# Defining the host is optional and defaults to https://api.e2b.dev
# See configuration.py for a list of all supported configuration parameters.
configuration = client.Configuration(
    host=API_HOST,
)


class E2BApiClient(client.ApiClient):
    def __init__(
        self,
        api_key: Union[str, None] = None,
        access_token: Union[str, None] = None,
        *args,
        **kwargs,
    ):
        configuration.api_key["ApiKeyAuth"] = api_key
        configuration.access_token = access_token

        super().__init__(configuration, *args, **kwargs)
        self.default_headers = default_headers


__all__ = ["E2BApiClient", "configuration", "client", "models"]



import warnings

from pathlib import PurePosixPath
from typing import Optional


def resolve_path(path: str, cwd: Optional[str] = None) -> str:
    warning = f"Path starts with {{0}} and cwd isn't set. The path {path} will evaluate to `{{1}}`, which may not be what you want."
    if path.startswith("./"):
        result = PurePosixPath(cwd or "/home/user", path).as_posix()
        if not cwd:
            warnings.warn(warning.format("./", result))
        return result

    if path.startswith("../"):
        result = PurePosixPath(cwd or "/home/user", path).as_posix()
        if not cwd:
            warnings.warn(warning.format("../", result))
        return result

    if path.startswith("~/"):
        result = PurePosixPath("/home/user", path[2:]).as_posix()
        warnings.warn(warning.format("~/", result))
        return result

    if not path.startswith("/") and cwd:
        return PurePosixPath(cwd, path).as_posix()

    return path



import json
import logging

from typing import (
    Generic,
    List,
    TYPE_CHECKING,
    TypeVar,
)

from e2b.sandbox.main import Sandbox

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from openai.types.beta.threads.run_submit_tool_outputs_params import ToolOutput
    from openai.types.beta.threads.run import Run


S = TypeVar(
    "S",
    bound=Sandbox,
)


class Actions(Generic[S]):
    def __init__(self, sandbox: S):
        self._sandbox = sandbox

    def run(self, run: "Run") -> List["ToolOutput"]:
        """
        Call the required actions for the provided run and return their outputs.

        :param run: OpenAI run object from `openai.beta.threads.runs.retrieve` or `openai.beta.threads.runs.retrieve.create` call that contains the names of the required actions and their arguments.

        :return: The outputs of the required actions in the run.
        """
        if run.status != "requires_action":
            return []

        if not run.required_action:
            return []

        outputs = []

        for tool_call in run.required_action.submit_tool_outputs.tool_calls:
            action = self._sandbox._actions.get(tool_call.function.name)
            if not action:
                logger.warning(f"Action {tool_call.function.name} not found.")
                continue

            args = json.loads(tool_call.function.arguments)
            output = action(self._sandbox, args)

            outputs.append(
                {
                    "tool_call_id": tool_call.id,
                    "output": output,
                }
            )

        return outputs


class OpenAI(Generic[S]):
    def __init__(self, actions: Actions[S]):
        self._actions = actions

    @property
    def actions(self):
        return self._actions



from os import getenv
from typing import Literal, Union, Optional

from . import Sandbox
from .exception import UnsupportedRuntimeException

CodeRuntime = Literal[
    "Node16",
    "Python3",
    "Bash",
    "Python3-DataAnalysis",
]


def run_code(
    runtime: Union[CodeRuntime, str],
    code: str,
    api_key: Optional[str] = None,
):
    """
    Run code in a sandboxed cloud playground and return the stdout and stderr.

    `run_code` wraps the `Sandbox` class and provides a simple interface for running code in a sandbox
    without any need to manage lifecycle of the sandbox.
    `run_code` automatically loads the E2B API key from the `E2B_API_KEY` environment variable.

    :param runtime: The runtime to run the code in. One of "Node16" or "Python3".
    :param code: The code to run
    :param api_key: The E2B API key to use. If not provided, the `E2B_API_KEY` environment variable is used.

    :return: A string touple of stdout and stderr
    """

    if api_key is None:
        api_key = getenv("E2B_API_KEY")

    if runtime == "Node16":
        template_id = "base"
        binary = "node"
        filepath = "/index.js"
    elif runtime == "Python3":
        template_id = "base"
        binary = "python3"
        filepath = "/index.py"
    elif runtime == "Python3-DataAnalysis":
        template_id = "Python3-DataAnalysis"
        binary = "python3"
        filepath = "/index.py"
    elif runtime == "Bash":
        template_id = "base"
        binary = "bash"
        filepath = "/main.sh"
    else:
        raise UnsupportedRuntimeException(
            f'Invalid runtime "{runtime}". Please contact us (hello@e2b.dev) if you need support for this runtime'
        )

    sandbox = Sandbox(template=template_id, api_key=api_key)
    sandbox.filesystem.write(filepath, code)

    output = sandbox.process.start_and_wait(cmd=f"{binary} {filepath}")

    sandbox.close()

    return output.stdout, output.stderr



from __future__ import annotations

import asyncio
import logging

from queue import Queue
from threading import Event
from typing import Any, Callable, List, Optional

from websockets.legacy.client import WebSocketClientProtocol, connect
from websockets.exceptions import ConnectionClosed
from websockets.typing import Data

logger = logging.getLogger(__name__)


class WebSocket:
    def __init__(
        self,
        url: str,
        started: Event,
        stopped: Event,
        queue_in: Queue[dict],
        queue_out: Queue[Data],
    ):
        self._ws: Optional[WebSocketClientProtocol] = None
        self.url = url
        self.started = started
        self.stopped = stopped
        self._process_cleanup: List[Callable[[], Any]] = []
        self._queue_in = queue_in
        self._queue_out = queue_out

    def run(self):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        loop.run_until_complete(self.async_run())

    async def async_run(self):
        await self._connect()
        await self.close()

    async def _send_message(self):
        logger.debug("WebSocket starting to send messages")
        while True:
            if self._queue_in.empty():
                await asyncio.sleep(0)
                continue
            message = self._queue_in.get()
            logger.debug(f"WebSocket message to send: {message}")
            if self._ws:
                await self._ws.send(message)
                logger.debug(f"WebSocket message sent: {message}")
                self._queue_in.task_done()
            else:
                logger.error("No WebSocket connection")

    async def _receive_message(self):
        try:
            if not self._ws:
                logger.error("No WebSocket connection")
                return
            async for message in self._ws:
                logger.debug(f"WebSocket received message: {message}".strip())
                self._queue_out.put(message)
        except Exception as e:
            logger.error(f"WebSocket received error while receiving messages: {e}")

    async def _connect(self):
        logger.debug(f"WebSocket connecting to {self.url}")

        ws_logger = logger.getChild("websockets.client")
        ws_logger.setLevel(logging.ERROR)

        websocket_connector = connect(
            self.url,
            max_size=None,
            max_queue=None,
            logger=ws_logger,
        )

        websocket_connector.BACKOFF_MIN = 1
        websocket_connector.BACKOFF_FACTOR = 1
        websocket_connector.BACKOFF_INITIAL = 0.2  # type: ignore

        async for websocket in websocket_connector:
            try:
                self._ws = websocket
                self.started.set()
                logger.info(f"WebSocket connected to {self.url}")

                send_task = asyncio.create_task(
                    self._send_message(), name="send_message"
                )
                self._process_cleanup.append(send_task.cancel)

                receive_task = asyncio.create_task(
                    self._receive_message(), name="receive_message"
                )
                self._process_cleanup.append(receive_task.cancel)

                while not self.stopped.is_set():
                    await asyncio.sleep(0)

                logger.info("WebSocket stopped")
                break
            except ConnectionClosed:
                logger.warning("WebSocket disconnected, it will try to reconnect")
                if self.stopped.is_set():
                    break

    async def close(self):
        for cancel in self._process_cleanup:
            cancel()

        self._process_cleanup.clear()

        if self._ws:
            await self._ws.close()



import base64
import logging
from typing import Any, List, Optional

from pydantic import BaseModel

from e2b.constants import TIMEOUT
from e2b.sandbox.exception import FilesystemException, RpcException
from e2b.sandbox.filesystem_watcher import FilesystemWatcher
from e2b.sandbox.sandbox_connection import SandboxConnection
from e2b.utils.filesystem import resolve_path

logger = logging.getLogger(__name__)


class FileInfo(BaseModel):
    """
    Information about a file or a directory in the sandbox.
    """

    is_dir: bool
    name: str


class FilesystemManager:
    """
    Manager for interacting with the filesystem in the sandbox.
    """

    _service_name = "filesystem"

    def __init__(self, sandbox: SandboxConnection):
        self._sandbox = sandbox

    @property
    def cwd(self) -> Optional[str]:
        return self._sandbox.cwd

    def read_bytes(self, path: str, timeout: Optional[float] = TIMEOUT) -> bytes:
        """
        Read the whole content of a file as a byte array.
        This can be used when you cannot represent the data as an UTF-8 string.

        :param path: path to a file
        :param timeout: timeout for the call
        :return: byte array representing the content of a file
        """
        path = resolve_path(path, self.cwd)
        result: str = self._sandbox._call(
            self._service_name, "readBase64", [path], timeout=timeout
        )
        return base64.b64decode(result)

    def write_bytes(
        self, path: str, content: bytes, timeout: Optional[float] = TIMEOUT
    ) -> None:
        """
        Write content to a file as a byte array.
        This can be used when you cannot represent the data as an UTF-8 string.

        A new file will be created if it doesn't exist.
        If the file already exists, it will be overwritten.

        :param path: path to a file
        :param timeout: timeout for the call
        :param content: byte array representing the content to write
        """
        path = resolve_path(path, self.cwd)
        base64_content = base64.b64encode(content).decode("utf-8")
        self._sandbox._call(
            self._service_name, "writeBase64", [path, base64_content], timeout=timeout
        )

    def read(self, path: str, timeout: Optional[float] = TIMEOUT) -> str:
        """
        Read the whole content of a file as an array of bytes.

        :param path: Path to a file
        :param timeout: Specify the duration, in seconds to give the method to finish its execution before it times out (default is 60 seconds). If set to None, the method will continue to wait until it completes, regardless of time
        :return: Content of a file
        """
        logger.debug(f"Reading file {path}")

        path = resolve_path(path, self.cwd)
        try:
            result: str = self._sandbox._call(
                self._service_name, "read", [path], timeout=timeout
            )
            logger.debug(f"Read file {path}")
            return result
        except RpcException as e:
            raise FilesystemException(e.message) from e

    def write(
        self, path: str, content: str, timeout: Optional[float] = TIMEOUT
    ) -> None:
        """
        Write content to a file.

        A new file will be created if it doesn't exist.
        If the file already exists, it will be overwritten.

        :param path: Path to a file
        :param content: Content to write
        :param timeout: Specify the duration, in seconds to give the method to finish its execution before it times out (default is 60 seconds). If set to None, the method will continue to wait until it completes, regardless of time
        """
        logger.debug(f"Writing file {path}")

        path = resolve_path(path, self.cwd)
        try:
            self._sandbox._call(
                self._service_name, "write", [path, content], timeout=timeout
            )
            logger.debug(f"Wrote file {path}")
        except RpcException as e:
            raise FilesystemException(e.message) from e

    def remove(self, path: str, timeout: Optional[float] = TIMEOUT) -> None:
        """
        Remove a file or a directory.

        :param path: Path to a file or a directory
        :param timeout: Specify the duration, in seconds to give the method to finish its execution before it times out (default is 60 seconds). If set to None, the method will continue to wait until it completes, regardless of time
        """
        logger.debug(f"Removing file {path}")

        path = resolve_path(path, self.cwd)
        try:
            self._sandbox._call(self._service_name, "remove", [path], timeout=timeout)
            logger.debug(f"Removed file {path}")
        except RpcException as e:
            raise FilesystemException(e.message) from e

    def list(self, path: str, timeout: Optional[float] = TIMEOUT) -> List[FileInfo]:
        """
        List files in a directory.

        :param path: Path to a directory
        :param timeout: Specify the duration, in seconds to give the method to finish its execution before it times out (default is 60 seconds). If set to None, the method will continue to wait until it completes, regardless of time

        :return: Array of files in a directory
        """
        logger.debug(f"Listing files in {path}")

        path = resolve_path(path, self.cwd)
        try:
            result: List[Any] = self._sandbox._call(
                self._service_name, "list", [path], timeout=timeout
            )
            logger.debug(f"Listed files in {path}, result: {result}")
            return [
                FileInfo(is_dir=file_info["isDir"], name=file_info["name"])
                for file_info in result
            ]
        except RpcException as e:
            raise FilesystemException(e.message) from e

    def make_dir(self, path: str, timeout: Optional[float] = TIMEOUT) -> None:
        """
        Create a new directory and all directories along the way if needed on the specified path.

        :param path: Path to a new directory. For example '/dirA/dirB' when creating 'dirB'
        :param timeout: Specify the duration, in seconds to give the method to finish its execution before it times out (default is 60 seconds). If set to None, the method will continue to wait until it completes, regardless of time
        """
        logger.debug(f"Creating directory {path}")

        path = resolve_path(path, self.cwd)
        try:
            self._sandbox._call(self._service_name, "makeDir", [path], timeout=timeout)
            logger.debug(f"Created directory {path}")
        except RpcException as e:
            raise FilesystemException(e.message) from e

    def watch_dir(self, path: str) -> FilesystemWatcher:
        """
        Watches directory for filesystem events.

        :param path: Path to a directory that will be watched

        :return: New watcher
        """
        logger.debug(f"Watching directory {path}")

        path = resolve_path(path, self.cwd)
        return FilesystemWatcher(
            connection=self._sandbox,
            path=path,
            service_name=self._service_name,
        )



import asyncio
from asyncio import AbstractEventLoop

from concurrent.futures import Future, CancelledError
from typing import Any, Awaitable, Callable, Generic, List, Optional, TypeVar

T = TypeVar("T")


# Check if using event is not better for most use cases
class DeferredFuture(Generic[T]):
    def __init__(self, cleanup_list: Optional[List[Callable[[], Any]]] = None):
        self._future = Future()
        if cleanup_list is not None:
            cleanup_list.append(self.cancel)

    def __call__(self, result: T):
        if not self._future.done():
            self._future.set_result(result)

    def result(self, timeout: Optional[float] = None) -> T:
        try:
            return self._future.result(timeout=timeout)
        # The thread was interrupted while waiting, it's okay to release the thread
        except CancelledError:
            pass

    def done(self) -> bool:
        return self._future.done()

    def cancel(self):
        if not self._future.done():
            self._future.cancel()

    def reject(self, reason: Exception):
        if not self._future.done():
            self._future.set_exception(reason)


def run_async_func_in_loop(loop: AbstractEventLoop, coro: Awaitable):
    asyncio.set_event_loop(loop)

    try:
        return loop.run_until_complete(coro)
    finally:
        loop.close()



from .code_snippet import OpenPort
from .env_vars import EnvVars
from .exception import (
    CurrentWorkingDirectoryDoesntExistException,
    FilesystemException,
    ProcessException,
    RpcException,
    SandboxException,
    TerminalException,
)
from .filesystem import FileInfo, FilesystemManager
from .filesystem_watcher import (
    FilesystemEvent,
    FilesystemOperation,
    FilesystemWatcher,
)
from .main import Sandbox, Action
from .process import (
    Process,
    ProcessManager,
    ProcessMessage,
    ProcessOutput,
)
from .run_code import run_code
from .terminal import (
    Terminal,
    TerminalManager,
    TerminalOutput,
)
from .sandbox_connection import RunningSandbox



import re


def camel_case_to_snake_case(text: str) -> str:
    return re.sub(r"(?<!^)(?=[A-Z])", "_", text).lower()


def snake_case_to_camel_case(text: str) -> str:
    string_split = text.split("_")
    return string_split[0] + "".join(word.capitalize() for word in string_split[1:])



from __future__ import annotations


import logging
import threading
from concurrent.futures import Future
from typing import Any, Callable, Optional

from pydantic import BaseModel

from e2b.constants import TIMEOUT
from e2b.sandbox.env_vars import EnvVars
from e2b.sandbox.exception import MultipleExceptions, RpcException, TerminalException
from e2b.sandbox.sandbox_connection import SandboxConnection, SubscriptionArgs
from e2b.utils.id import create_id

logger = logging.getLogger(__file__)


class TerminalOutput(BaseModel):
    data: str = ""

    def _add_data(self, data: str) -> None:
        self.data += data


class Terminal:
    """
    Terminal session.
    """

    @property
    def data(self) -> str:
        """
        Terminal output data.
        """
        return self._output.data

    @property
    def output(self) -> TerminalOutput:
        """
        Terminal output.
        """
        return self._output

    @property
    def finished(self):
        """
        A future that is resolved when the terminal session exits.
        """
        return self._finished

    @property
    def terminal_id(self) -> str:
        """
        The terminal id used to identify the terminal in the session.
        """
        return self._terminal_id

    def wait(self):
        """
        Wait till the terminal session exits.
        """
        return self.finished.result()

    def __init__(
        self,
        terminal_id: str,
        sandbox: SandboxConnection,
        trigger_exit: Callable[[], Any],
        finished: Future[TerminalOutput],
        output: TerminalOutput,
    ):
        self._terminal_id = terminal_id
        self._sandbox = sandbox
        self._trigger_exit = trigger_exit
        self._finished = finished
        self._output = output

    def send_data(self, data: str, timeout: Optional[float] = TIMEOUT) -> None:
        """
        Send data to the terminal standard input.

        :param data: Data to send
        :param timeout: Specify the duration, in seconds to give the method to finish its execution before it times out (default is 60 seconds). If set to None, the method will continue to wait until it completes, regardless of time
        """
        try:
            self._sandbox._call(
                TerminalManager._service_name,
                "data",
                [self.terminal_id, data],
                timeout=timeout,
            )
        except RpcException as e:
            raise TerminalException(e.message) from e

    def resize(self, cols: int, rows: int, timeout: Optional[float] = TIMEOUT) -> None:
        """
        Resizes the terminal tty.

        :param cols: Number of columns
        :param rows: Number of rows
        :param timeout: Specify the duration, in seconds to give the method to finish its execution before it times out (default is 60 seconds). If set to None, the method will continue to wait until it completes, regardless of time
        """
        try:
            self._sandbox._call(
                TerminalManager._service_name,
                "resize",
                [self.terminal_id, cols, rows],
                timeout=timeout,
            )
        except RpcException as e:
            raise TerminalException(e.message) from e

    def kill(self, timeout: Optional[float] = TIMEOUT) -> None:
        """
        Kill the terminal session.

        :param timeout: Specify the duration, in seconds to give the method to finish its execution before it times out (default is 60 seconds). If set to None, the method will continue to wait until it completes, regardless of time
        """
        try:
            self._sandbox._call(
                TerminalManager._service_name,
                "destroy",
                [self.terminal_id],
                timeout=timeout,
            )
        except RpcException as e:
            raise TerminalException(e.message) from e
        self._trigger_exit()


class TerminalManager:
    """
    Manager for starting and interacting with terminal sessions in the sandbox.
    """

    _service_name = "terminal"

    def __init__(self, sandbox: SandboxConnection):
        self._sandbox = sandbox

    def start(
        self,
        on_data: Callable[[str], Any],
        cols: int,
        rows: int,
        cwd: str = "",
        terminal_id: Optional[str] = None,
        on_exit: Optional[Callable[[], Any]] = None,
        cmd: Optional[str] = None,
        env_vars: Optional[EnvVars] = None,
        timeout: Optional[float] = TIMEOUT,
    ) -> Terminal:
        """
        Start a new terminal session.

        :param on_data: Callback that will be called when the terminal sends data
        :param cwd: Working directory where will the terminal start
        :param terminal_id: Unique identifier of the terminal session
        :param on_exit: Callback that will be called when the terminal exits
        :param cols: Number of columns the terminal will have. This affects rendering
        :param rows: Number of rows the terminal will have. This affects rendering
        :param cmd: If the `cmd` parameter is defined it will be executed as a command
        and this terminal session will exit when the command exits
        :param env_vars: Environment variables that will be accessible inside of the terminal
        :param timeout: Specify the duration, in seconds to give the method to finish its execution before it times out (default is 60 seconds). If set to None, the method will continue to wait until it completes, regardless of time

        :return: Terminal session
        """
        env_vars = self._sandbox.env_vars.update(env_vars or {})

        future_exit = Future()
        terminal_id = terminal_id or create_id(12)

        unsub_all: Optional[Callable[[], Any]] = None

        output = TerminalOutput()

        def handle_data(data: str):
            output._add_data(data)
            on_data(data)

        try:
            unsub_all = self._sandbox._handle_subscriptions(
                SubscriptionArgs(
                    service=self._service_name,
                    handler=handle_data,
                    method="onData",
                    params=[terminal_id],
                ),
                SubscriptionArgs(
                    service=self._service_name,
                    handler=lambda result: future_exit.set_result(result),
                    method="onExit",
                    params=[terminal_id],
                ),
            )
        except (RpcException, MultipleExceptions) as e:
            future_exit.cancel()

            if isinstance(e, RpcException):
                raise TerminalException(e.message) from e
            elif isinstance(e, MultipleExceptions):
                raise TerminalException(
                    "Failed to subscribe to RPC services necessary for starting terminal"
                ) from e

        # TODO: Handle exit handler finish for exits (the same for processes)
        future_exit_handler_finish: Future[TerminalOutput] = Future()

        def exit_handler():
            future_exit.result()

            if unsub_all:
                unsub_all()

            if on_exit:
                on_exit()
            future_exit_handler_finish.set_result(output)

        threading.Thread(
            name="e2b-terminal-exit-handler", daemon=True, target=exit_handler
        ).start()

        def trigger_exit():
            future_exit.set_result(None)
            future_exit_handler_finish.result()

        try:
            if not cwd and self._sandbox.cwd:
                cwd = self._sandbox.cwd

            self._sandbox._call(
                self._service_name,
                "start",
                [
                    terminal_id,
                    cols,
                    rows,
                    env_vars if env_vars else {},
                    cmd,
                    cwd,
                ],
                timeout=timeout,
            )
            return Terminal(
                terminal_id=terminal_id,
                sandbox=self._sandbox,
                trigger_exit=trigger_exit,
                finished=future_exit_handler_finish,
                output=output,
            )
        except RpcException as e:
            trigger_exit()
            raise TerminalException(e.message) from e



from enum import Enum
from pydantic import BaseModel


class OutType(str, Enum):
    Stdout = "Stdout"
    Stderr = "Stderr"


class OutResponse(BaseModel):
    type: OutType
    timestamp: int
    """
    Unix epoch in nanoseconds.
    """
    line: str


class OutStdoutResponse(OutResponse):
    type: OutType = OutType.Stdout


class OutStderrResponse(OutResponse):
    type: OutType = OutType.Stderr



from typing import Dict, List, Optional


class SandboxException(Exception):
    pass


class SandboxNotOpenException(Exception):
    pass


class RpcException(SandboxException):
    def __init__(
        self,
        message: str,
        code: int,
        id: str,
        data: Optional[Dict] = None,
    ):
        super().__init__(message)
        self.data = data
        self.code = code
        self.message = message
        self.id = id


class MultipleExceptions(SandboxException):
    def __init__(self, message: str, exceptions: List[Exception]):
        super().__init__(f"Multiple exceptions occurred: {message}")
        self.exceptions = exceptions


class FilesystemException(SandboxException):
    pass


class ProcessException(SandboxException):
    pass


class CurrentWorkingDirectoryDoesntExistException(ProcessException):
    pass


class TerminalException(SandboxException):
    pass


class AuthenticationException(SandboxException):
    pass


class UnsupportedRuntimeException(SandboxException):
    pass


class TimeoutException(SandboxException):
    pass



from __future__ import annotations

import logging
import re
import inspect
import threading
import time
from concurrent.futures import Future
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    List,
    Optional,
    Union,
)
from pydantic import BaseModel

from e2b.constants import TIMEOUT
from e2b.sandbox.env_vars import EnvVars
from e2b.sandbox.exception import (
    MultipleExceptions,
    ProcessException,
    RpcException,
    CurrentWorkingDirectoryDoesntExistException,
)
from e2b.sandbox.out import OutStderrResponse, OutStdoutResponse
from e2b.sandbox.sandbox_connection import SandboxConnection, SubscriptionArgs
from e2b.utils.id import create_id

logger = logging.getLogger(__name__)


class ProcessMessage(BaseModel):
    """
    A message from a process.
    """

    line: str
    error: bool = False
    timestamp: int
    """
    Unix epoch in nanoseconds
    """

    def __str__(self):
        return self.line


class ProcessOutput(BaseModel):
    """
    Output from a process.
    """

    delimiter: ClassVar[str] = "\n"
    messages: List[ProcessMessage] = []

    error: bool = False
    exit_code: Optional[int] = None

    @property
    def stdout(self) -> str:
        """
        The stdout from the process.
        """
        return self.delimiter.join(out.line for out in self.messages if not out.error)

    @property
    def stderr(self) -> str:
        """
        The stderr from the process.
        """
        return self.delimiter.join(out.line for out in self.messages if out.error)

    def _insert_by_timestamp(self, message: ProcessMessage):
        """Insert an out based on its timestamp using insertion sort."""
        i = len(self.messages) - 1
        while i >= 0 and self.messages[i].timestamp > message.timestamp:
            i -= 1
        self.messages.insert(i + 1, message)

    def _add_stdout(self, message: ProcessMessage):
        self._insert_by_timestamp(message)

    def _add_stderr(self, message: ProcessMessage):
        self.error = True
        self._insert_by_timestamp(message)


class Process:
    """
    A process running in the sandbox.
    """

    def __init__(
        self,
        process_id: str,
        sandbox: SandboxConnection,
        trigger_exit: Callable[[], Any],
        finished: Future[ProcessOutput],
        output: ProcessOutput,
    ):
        self._process_id = process_id
        self._sandbox = sandbox
        self._trigger_exit = trigger_exit
        self._finished = finished
        self._output = output

    @property
    def exit_code(self) -> Optional[int]:
        """
        The exit code of the last process started by this manager.
        """
        if not self.finished:
            raise ProcessException("Process has not finished yet")
        return self.output.exit_code

    @property
    def output(self) -> ProcessOutput:
        """
        The output from the process.
        """
        return self._output

    @property
    def stdout(self) -> str:
        """
        The stdout from the process.
        """
        return self._output.stdout

    @property
    def stderr(self) -> str:
        """
        The stderr from the process.
        """
        return self._output.stderr

    @property
    def error(self) -> bool:
        """
        True if the process has written to stderr.
        """
        return self._output.error

    @property
    def output_messages(self) -> List[ProcessMessage]:
        """
        The output messages from the process.
        """
        return self._output.messages

    @property
    def finished(self):
        """
        A future that is resolved when the process exits.
        """
        return self._finished

    @property
    def process_id(self) -> str:
        """
        The process id used to identify the process in the sandbox.
        This is not the system process id of the process running in the sandbox.
        """
        return self._process_id

    def wait(self, timeout: Optional[float] = None) -> ProcessOutput:
        """
        Wait for the process to exit.

        :param timeout: Specify the duration, in seconds to give the method to finish its execution before it times out. If set to None, the method will continue to wait until it completes, regardless of time
        """
        start_time = time.time()
        while timeout is None or time.time() - start_time < timeout:
            if self._finished.done():
                return self._finished.result()
            time.sleep(0.1)
            if not self._sandbox.is_open:
                break

    def send_stdin(self, data: str, timeout: Optional[float] = TIMEOUT) -> None:
        """
        Send data to the process stdin.

        :param data: Data to send
        :param timeout: Specify the duration, in seconds to give the method to finish its execution before it times out (default is 60 seconds). If set to None, the method will continue to wait until it completes, regardless of time
        """
        try:
            self._sandbox._call(
                ProcessManager._service_name,
                "stdin",
                [self.process_id, data],
                timeout=timeout,
            )
        except RpcException as e:
            raise ProcessException(e.message) from e

    def kill(self, timeout: Optional[float] = TIMEOUT) -> None:
        """
        Kill the process.

        :param timeout: Specify the duration, in seconds to give the method to finish its execution before it times out (default is 60 seconds). If set to None, the method will continue to wait until it completes, regardless of time
        """
        try:
            self._sandbox._call(
                ProcessManager._service_name, "kill", [self.process_id], timeout=timeout
            )
        except RpcException as e:
            raise ProcessException(e.message) from e
        self._trigger_exit()


class ProcessManager:
    """
    Manager for starting and interacting with processes in the sandbox.
    """

    _service_name = "process"

    def __init__(
        self,
        sandbox: SandboxConnection,
        on_stdout: Optional[Callable[[ProcessMessage], Any]] = None,
        on_stderr: Optional[Callable[[ProcessMessage], Any]] = None,
        on_exit: Optional[Union[Callable[[int], Any], Callable[[], Any]]] = None,
    ):
        self._sandbox = sandbox
        self._process_cleanup: List[Callable[[], Any]] = []
        self._on_stdout = on_stdout
        self._on_stderr = on_stderr
        self._on_exit = on_exit

    def start(
        self,
        cmd: str,
        on_stdout: Optional[Callable[[ProcessMessage], Any]] = None,
        on_stderr: Optional[Callable[[ProcessMessage], Any]] = None,
        on_exit: Optional[Union[Callable[[int], Any], Callable[[], Any]]] = None,
        env_vars: Optional[EnvVars] = None,
        cwd: str = "",
        rootdir: str = "",  # DEPRECATED
        process_id: Optional[str] = None,
        timeout: Optional[float] = TIMEOUT,
    ) -> Process:
        logger.info(f"Starting process: {cmd}")
        env_vars = env_vars or {}
        env_vars = {**self._sandbox.env_vars, **env_vars}

        on_stdout = on_stdout or self._on_stdout
        on_stderr = on_stderr or self._on_stderr
        on_exit = on_exit or self._on_exit

        future_exit = Future()
        process_id = process_id or create_id(12)

        unsub_all: Optional[Callable] = None

        output = ProcessOutput()

        def handle_exit(exit_code: int):
            output.exit_code = exit_code
            logger.info(f"Process {process_id} exited with exit code {exit_code}")
            future_exit.set_result(True)

        def handle_stdout(data: Dict[Any, Any]):
            out = OutStdoutResponse(**data)

            message = ProcessMessage(
                line=out.line,
                timestamp=out.timestamp,
                error=False,
            )

            output._add_stdout(message)
            if on_stdout:
                try:
                    on_stdout(message)
                except TypeError as error:
                    logger.exception(f"Error in on_stdout callback: {error}")

        def handle_stderr(data: Dict[Any, Any]):
            out = OutStderrResponse(**data)

            message = ProcessMessage(
                line=out.line,
                timestamp=out.timestamp,
                error=True,
            )

            output._add_stderr(message)
            if on_stderr:
                try:
                    on_stderr(message)
                except TypeError as error:
                    logger.exception(f"Error in on_stdout callback: {error}")

        try:
            subscription_args = [
                SubscriptionArgs(
                    service=self._service_name,
                    handler=handle_exit,
                    method="onExit",
                    params=[process_id],
                ),
                SubscriptionArgs(
                    service=self._service_name,
                    handler=handle_stdout,
                    method="onStdout",
                    params=[process_id],
                ),
                SubscriptionArgs(
                    service=self._service_name,
                    handler=handle_stderr,
                    method="onStderr",
                    params=[process_id],
                ),
            ]
            unsub_all = self._sandbox._handle_subscriptions(*subscription_args)

        except (RpcException, MultipleExceptions) as e:
            future_exit.cancel()

            if isinstance(e, RpcException):
                raise ProcessException(e.message) from e
            elif isinstance(e, MultipleExceptions):
                raise ProcessException(
                    "Failed to subscribe to RPC services necessary for starting process"
                ) from e

        future_exit_handler_finish: Future[ProcessOutput] = Future()

        def exit_handler():
            future_exit.result()
            logger.info(f"Handling process exit (id: {process_id})")
            if unsub_all:
                unsub_all()
            if on_exit:
                sig = inspect.signature(on_exit)
                params = sig.parameters.values()
                try:
                    if len(params) == 0:
                        on_exit()
                    else:
                        on_exit(output.exit_code or 0)
                except TypeError as error:
                    logger.exception(f"Error in on_exit callback: {error}")
            future_exit_handler_finish.set_result(output)

        threading.Thread(
            name="e2b-process-exit-handler", daemon=True, target=exit_handler
        ).start()

        def trigger_exit():
            logger.info(f"Exiting the process (id: {process_id})")
            future_exit.set_result(None)
            future_exit_handler_finish.result()
            logger.debug(f"Exited the process (id: {process_id})")

        try:
            if not cwd and rootdir:
                cwd = rootdir
                logger.warning("The rootdir parameter is deprecated, use cwd instead.")

            if not cwd and self._sandbox.cwd:
                cwd = self._sandbox.cwd

            self._sandbox._call(
                self._service_name,
                "start",
                [
                    process_id,
                    cmd,
                    env_vars,
                    cwd,
                ],
                timeout=timeout,
            )
            logger.info(f"Started process (id: {process_id})")
            return Process(
                output=output,
                sandbox=self._sandbox,
                process_id=process_id,
                trigger_exit=trigger_exit,
                finished=future_exit_handler_finish,
            )
        except RpcException as e:
            trigger_exit()
            if re.match(
                r"error starting process '\w+': fork/exec /bin/bash: no such file or directory",
                e.message,
            ):
                raise CurrentWorkingDirectoryDoesntExistException(
                    "Failed to start the process. You are trying set `cwd` to a directory that does not exist."
                ) from e
            raise ProcessException(e.message) from e
        except TimeoutError as e:
            logger.error(f"Timeout error during starting the process: {cmd}")
            trigger_exit()
            raise e

    def start_and_wait(
        self,
        cmd: str,
        on_stdout: Optional[Callable[[ProcessMessage], Any]] = None,
        on_stderr: Optional[Callable[[ProcessMessage], Any]] = None,
        on_exit: Optional[Callable[[int], Any]] = None,
        env_vars: Optional[EnvVars] = None,
        cwd: str = "",
        process_id: Optional[str] = None,
        timeout: Optional[float] = TIMEOUT,
    ) -> ProcessOutput:
        return self.start(
            cmd,
            on_stdout=on_stdout,
            on_stderr=on_stderr,
            on_exit=on_exit,
            env_vars=env_vars,
            cwd=cwd,
            process_id=process_id,
            timeout=timeout,
        ).wait()



# coding: utf-8

"""
    E2B API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: 0.1.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json


from typing import Any, Dict, Optional
from pydantic import BaseModel, Field, StrictStr


class NewInstance(BaseModel):
    """
    NewInstance
    """

    env_id: StrictStr = Field(
        ..., alias="envID", description="Identifier of the required environment"
    )
    metadata: Optional[Dict[str, StrictStr]] = None
    additional_properties: Dict[str, Any] = {}
    __properties = ["envID", "metadata"]

    class Config:
        """Pydantic configuration"""

        allow_population_by_field_name = True
        validate_assignment = True

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> NewInstance:
        """Create an instance of NewInstance from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self):
        """Returns the dictionary representation of the model using alias"""
        _dict = self.dict(
            by_alias=True, exclude={"additional_properties"}, exclude_none=True
        )
        # puts key-value pairs in additional_properties in the top level
        if self.additional_properties is not None:
            for _key, _value in self.additional_properties.items():
                _dict[_key] = _value

        return _dict

    @classmethod
    def from_dict(cls, obj: dict) -> NewInstance:
        """Create an instance of NewInstance from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return NewInstance.parse_obj(obj)

        _obj = NewInstance.parse_obj(
            {"env_id": obj.get("envID"), "metadata": obj.get("metadata")}
        )
        # store additional fields in additional_properties
        for _key in obj.keys():
            if _key not in cls.__properties:
                _obj.additional_properties[_key] = obj.get(_key)

        return _obj



from __future__ import annotations

import asyncio
import json
import logging
import threading
import time

from concurrent.futures import ThreadPoolExecutor, TimeoutError
from queue import Queue
from threading import Event
from typing import Any, Callable, Dict, Iterator, List, Union, Optional
from jsonrpcclient import Error, Ok, request_json
from jsonrpcclient.id_generators import decimal as decimal_id_generator
from jsonrpcclient.responses import Response
from pydantic import BaseModel, PrivateAttr, ConfigDict
from websockets.typing import Data

from e2b.constants import TIMEOUT
from e2b.sandbox.exception import RpcException, TimeoutException, SandboxException
from e2b.sandbox.websocket_client import WebSocket
from e2b.utils.future import DeferredFuture, run_async_func_in_loop
from e2b.utils.threads import shutdown_executor

logger = logging.getLogger(__name__)


class Notification(BaseModel):
    """Nofification."""

    method: str
    params: Dict


Message = Union[Response, Notification]


def to_response_or_notification(response: Dict[str, Any]) -> Message:
    """Create a Response namedtuple from a dict."""
    if "error" in response:
        return Error(
            response["error"]["code"],
            response["error"]["message"],
            response["error"].get("data"),
            response["id"],
        )
    elif "result" in response and "id" in response:
        return Ok(response["result"], response["id"])

    elif "params" in response:
        return Notification(method=response["method"], params=response["params"])

    raise ValueError("Invalid response", response)


class SandboxRpc(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    url: str
    on_message: Callable[[Notification], None]

    _id_generator: Iterator[int] = PrivateAttr(default_factory=decimal_id_generator)
    _waiting_for_replies: Dict[int, DeferredFuture] = PrivateAttr(default_factory=dict)
    _queue_in: Queue = PrivateAttr(default_factory=Queue)
    _queue_out: Queue = PrivateAttr(default_factory=Queue)
    _process_cleanup: List[Callable[[], Any]] = PrivateAttr(default_factory=list)
    _websocket_task: Optional[threading.Thread] = PrivateAttr(default=None)
    _closed: bool = PrivateAttr(default=False)

    def process_messages(self):
        while True:
            data = self._queue_out.get()
            logger.debug(f"WebSocket received message: {data}".strip())
            self._receive_message(data)
            self._queue_out.task_done()

    def connect(self, timeout: Optional[float] = TIMEOUT):
        started = Event()
        stopped = Event()
        self._process_cleanup.append(stopped.set)

        threading.Thread(
            target=self.process_messages, daemon=True, name="e2b-process-messages"
        ).start()

        threading.Thread(
            target=WebSocket(
                url=self.url,
                queue_in=self._queue_in,
                queue_out=self._queue_out,
                started=started,
                stopped=stopped,
            ).run,
            daemon=True,
            name="e2b-websocket",
        ).start()

        logger.info("WebSocket waiting to start")

        try:
            start_time = time.time()
            while (
                not started.is_set()
                and time.time() - start_time < timeout
                and not self._closed
            ):
                time.sleep(0.1)

            if not started.is_set():
                logger.error("WebSocket failed to start")
                raise TimeoutException("WebSocket failed to start")
        except BaseException as e:
            self.close()
            raise SandboxException(f"WebSocket failed to start: {e}") from e

        logger.info("WebSocket started")

    def send_message(
        self,
        method: str,
        params: List[Any],
        timeout: Optional[float],
    ) -> Any:
        timeout = timeout or TIMEOUT

        id = next(self._id_generator)
        request = request_json(method, params, id)
        future_reply = DeferredFuture(self._process_cleanup)

        try:
            self._waiting_for_replies[id] = future_reply
            logger.debug(f"WebSocket queueing message: {request}")
            self._queue_in.put(request)
            logger.debug(f"WebSocket waiting for reply: {request}")
            try:
                r = future_reply.result(timeout=timeout)
            except TimeoutError as e:
                logger.error(f"WebSocket timed out while waiting for: {request} {e}")
                raise TimeoutException(
                    f"WebSocket timed out while waiting for: {request} {e}"
                )
            return r
        except Exception as e:
            logger.error(f"WebSocket received error while waiting for: {request} {e}")
            raise e
        finally:
            del self._waiting_for_replies[id]
            logger.debug(f"WebSocket removed waiting handler for {id}")

    def _receive_message(self, data: Data):
        logger.debug(f"Processing message: {data}".strip())

        message = to_response_or_notification(json.loads(data))

        logger.debug(
            f"Current waiting handlers: {list(self._waiting_for_replies.keys())}"
        )
        if isinstance(message, Ok):
            if (
                message.id in self._waiting_for_replies
                and self._waiting_for_replies[message.id]
            ):
                self._waiting_for_replies[message.id](message.result)
                return
        elif isinstance(message, Error):
            if (
                message.id in self._waiting_for_replies
                and self._waiting_for_replies[message.id]
            ):
                self._waiting_for_replies[message.id].reject(
                    RpcException(
                        code=message.code,
                        message=message.message,
                        id=message.id,
                        data=message.data,
                    )
                )
                return

        elif isinstance(message, Notification):
            self.on_message(message)

    def _close(self):
        self._closed = True

        for cancel in self._process_cleanup:
            cancel()

        self._process_cleanup.clear()

        for handler in self._waiting_for_replies.values():
            handler.cancel()
            del handler

    def close(self):
        self._close()



from os import getenv
from typing import Optional

from e2b.sandbox.exception import AuthenticationException


def get_api_key(api_key: Optional[str]) -> str:
    api_key = api_key or getenv("E2B_API_KEY")

    if api_key is None:
        raise AuthenticationException(
            "API key is required, please visit https://e2b.dev/docs to get your API key. "
            "You can either set the environment variable `E2B_API_KEY` "
            'or you can pass it directly to the sandbox like Sandbox(api_key="e2b_...")',
        )

    return api_key



import binascii
import json
import os
import time
import uuid
from time import sleep
from typing import Optional, Callable, Any, List, Union, Literal

import requests
from pydantic import BaseModel
from websocket import create_connection

from e2b import Sandbox, EnvVars, ProcessMessage, OpenPort
from e2b.constants import TIMEOUT


class Error(BaseModel):
    name: str
    value: str
    traceback: List[str]


class Result(BaseModel):
    output: Optional[str] = None
    stdout: List[str] = []
    stderr: List[str] = []
    error: Optional[Error] = None
    # TODO: This will be changed in the future, it's just to enable the use of display_data
    display_data: List[dict] = []


class CodeInterpreterV2(Sandbox):
    template = "code-interpreter-stateful"

    def __init__(
        self,
        template: Optional[str] = None,
        api_key: Optional[str] = None,
        cwd: Optional[str] = None,
        env_vars: Optional[EnvVars] = None,
        timeout: Optional[float] = TIMEOUT,
        on_stdout: Optional[Callable[[ProcessMessage], Any]] = None,
        on_stderr: Optional[Callable[[ProcessMessage], Any]] = None,
        on_exit: Optional[Callable[[int], Any]] = None,
        **kwargs,
    ):
        super().__init__(
            template=template or self.template,
            api_key=api_key,
            cwd=cwd,
            env_vars=env_vars,
            timeout=timeout,
            on_stdout=on_stdout,
            on_stderr=on_stderr,
            on_exit=on_exit,
            **kwargs,
        )

        self._setup_jupyter()

    def _setup_jupyter(self) -> None:
        config = json.loads(self.filesystem.read("/root/.jupyter/config.json"))
        self._jupyter_server_token = config["token"]
        self._jupyter_kernel_id = config["kernel_id"]

    def _connect_kernel(self):
        header = {"Authorization": f"Token {self._jupyter_server_token}"}
        return create_connection(
            f"{self.get_protocol('ws')}://{self.get_hostname(8888)}/api/kernels/{self._jupyter_kernel_id}/channels",
            header=header,
        )

    @staticmethod
    def _send_execute_request(code: str) -> dict:
        msg_id = str(uuid.uuid4())
        session = str(uuid.uuid4())

        return {
            "header": {
                "msg_id": msg_id,
                "username": "e2b",
                "session": session,
                "msg_type": "execute_request",
                "version": "5.3",
            },
            "parent_header": {},
            "metadata": {},
            "content": {
                "code": code,
                "silent": False,
                "store_history": False,
                "user_expressions": {},
                "allow_stdin": False,
            },
        }

    @staticmethod
    def _wait_for_result(
        ws,
        on_stdout: Optional[Callable[[ProcessMessage], Any]],
        on_stderr: Optional[Callable[[ProcessMessage], Any]],
    ) -> Result:
        result = Result()
        input_accepted = False

        while True:
            response = json.loads(ws.recv())
            if response["msg_type"] == "error":
                result.error = Error(
                    name=response["content"]["ename"],
                    value=response["content"]["evalue"],
                    traceback=response["content"]["traceback"],
                )

            elif response["msg_type"] == "stream":
                if response["content"]["name"] == "stdout":
                    result.stdout.append(response["content"]["text"])
                    if on_stdout:
                        on_stdout(
                            ProcessMessage(
                                line=response["content"]["text"],
                                timestamp=time.time_ns(),
                            )
                        )

                elif response["content"]["name"] == "stderr":
                    result.stderr.append(response["content"]["text"])
                    if on_stderr:
                        on_stderr(
                            ProcessMessage(
                                line=response["content"]["text"],
                                error=True,
                                timestamp=time.time_ns(),
                            )
                        )

            elif response["msg_type"] == "display_data":
                result.display_data.append(response["content"]["data"])

            elif response["msg_type"] == "execute_result":
                result.output = response["content"]["data"]["text/plain"]

            elif response["msg_type"] == "status":
                if response["content"]["execution_state"] == "idle":
                    if input_accepted:
                        break
                elif response["content"]["execution_state"] == "error":
                    result.error = Error(
                        name=response["content"]["ename"],
                        value=response["content"]["evalue"],
                        traceback=response["content"]["traceback"],
                    )
                    break

            elif response["msg_type"] == "execute_reply":
                if response["content"]["status"] == "error":
                    result.error = Error(
                        name=response["content"]["ename"],
                        value=response["content"]["evalue"],
                        traceback=response["content"]["traceback"],
                    )
                elif response["content"]["status"] == "ok":
                    pass

            elif response["msg_type"] == "execute_input":
                input_accepted = True
            else:
                print("[UNHANDLED MESSAGE TYPE]:", response["msg_type"])
        return result

    def exec_python(
        self,
        code: str,
        on_stdout: Optional[Callable[[ProcessMessage], Any]] = None,
        on_stderr: Optional[Callable[[ProcessMessage], Any]] = None,
    ) -> Result:
        ws = self._connect_kernel()
        ws.send(json.dumps(self._send_execute_request(code)))
        result = self._wait_for_result(ws, on_stdout, on_stderr)

        ws.close()

        return result



import logging
import time
from typing import Optional, Callable, Any, Tuple, List, Union
from e2b.sandbox.process import ProcessMessage

from pydantic import BaseModel, PrivateAttr

from e2b import EnvVars, Sandbox
from e2b.constants import TIMEOUT

logger = logging.getLogger(__name__)


class Artifact(BaseModel):
    name: str
    _sandbox: Sandbox = PrivateAttr()

    def __init__(self, sandbox: Sandbox, **data: Any):
        super().__init__(**data)
        self._sandbox = sandbox

    def __hash__(self):
        return hash(self.name)

    def download(self) -> bytes:
        return self._sandbox.download_file(self.name)


class DataAnalysis(Sandbox):
    template = "Python3-DataAnalysis"

    def __init__(
        self,
        template: Optional[str] = None,
        api_key: Optional[str] = None,
        cwd: Optional[str] = None,
        env_vars: Optional[EnvVars] = None,
        timeout: Optional[float] = TIMEOUT,
        on_stdout: Optional[Callable[[ProcessMessage], Any]] = None,
        on_stderr: Optional[Callable[[ProcessMessage], Any]] = None,
        on_artifact: Optional[Callable[[Artifact], Any]] = None,
        on_exit: Optional[Callable[[int], Any]] = None,
        **kwargs,
    ):
        self.on_artifact = on_artifact
        super().__init__(
            template=template or self.template,
            api_key=api_key,
            cwd=cwd,
            env_vars=env_vars,
            timeout=timeout,
            on_stdout=on_stdout,
            on_stderr=on_stderr,
            on_exit=on_exit,
            **kwargs,
        )

    def run_python(
        self,
        code: str,
        on_stdout: Optional[Callable[[ProcessMessage], Any]] = None,
        on_stderr: Optional[Callable[[ProcessMessage], Any]] = None,
        on_artifact: Optional[Callable[[Artifact], Any]] = None,
        on_exit: Optional[Callable[[int], Any]] = None,
        env_vars: Optional[EnvVars] = None,
        cwd: str = "",
        process_id: Optional[str] = None,
        timeout: Optional[float] = TIMEOUT,
    ) -> Tuple[str, str, List[Artifact]]:
        artifacts = set()

        def register_artifacts(event: Any) -> None:
            on_artifact_func = on_artifact or self.on_artifact
            if event.operation == "Create":
                artifact = Artifact(name=event.path, sandbox=self)
                artifacts.add(artifact)
                if on_artifact_func:
                    try:
                        on_artifact_func(artifact)
                    except Exception as e:
                        logger.error("Failed to process artifact", exc_info=e)

        watcher = self.filesystem.watch_dir("/home/user/artifacts")
        watcher.add_event_listener(register_artifacts)
        watcher.start()

        epoch_time = time.time()
        codefile_path = f"/tmp/main-{epoch_time}.py"
        self.filesystem.write(codefile_path, code)

        output = self.process.start_and_wait(
            f"python {codefile_path}",
            on_stdout=on_stdout,
            on_stderr=on_stderr,
            on_exit=on_exit,
            env_vars=env_vars,
            cwd=cwd,
            process_id=process_id,
            timeout=timeout,
        )

        watcher.stop()

        return output.stdout, output.stderr, list(artifacts)

    def _install_packages(
        self,
        command: str,
        package_names: Union[str, List[str]],
        timeout: Optional[float] = TIMEOUT,
    ) -> None:
        if isinstance(package_names, list):
            package_names = " ".join(package_names)

        package_names = package_names.strip()
        if not package_names:
            return

        output = self.process.start_and_wait(
            f"{command} {package_names}", timeout=timeout
        )

        if output.exit_code != 0:
            raise Exception(
                f"Failed to install package {package_names}: {output.stderr}"
            )

    def install_python_packages(
        self, package_names: Union[str, List[str]], timeout: Optional[float] = TIMEOUT
    ) -> None:
        self._install_packages("pip install", package_names, timeout=timeout)

    def install_system_packages(
        self, package_names: Union[str, List[str]], timeout: Optional[float] = TIMEOUT
    ) -> None:
        self._install_packages(
            "sudo apt-get install -y", package_names, timeout=timeout
        )


CodeInterpreter = DataAnalysis



import logging
import urllib.parse
import requests
import threading

from os import path
from typing import Any, Callable, Dict, List, Literal, Optional, IO, TypeVar, Union
from typing_extensions import Self

from e2b.api import models
from e2b.constants import TIMEOUT, ENVD_PORT, FILE_ROUTE
from e2b.sandbox.code_snippet import CodeSnippetManager, OpenPort
from e2b.sandbox.env_vars import EnvVars
from e2b.sandbox.filesystem import FilesystemManager
from e2b.sandbox.process import ProcessManager, ProcessMessage
from e2b.sandbox.sandbox_connection import SandboxConnection
from e2b.sandbox.terminal import TerminalManager

logger = logging.getLogger(__name__)


S = TypeVar(
    "S",
    bound="Sandbox",
)

Action = Callable[[S, Dict[str, Any]], str]


class Sandbox(SandboxConnection):
    """
    E2B cloud sandbox gives your agent a full cloud development environment that's sandboxed.

    That means:
    - Access to Linux OS
    - Using filesystem (create, list, and delete files and dirs)
    - Run processes
    - Sandboxed - you can run any code
    - Access to the internet

    Check usage docs - https://e2b.dev/docs/sandbox/overview

    These cloud sandboxes are meant to be used for agents. Like a sandboxed playgrounds, where the agent can do whatever it wants.
    """

    @property
    def process(self) -> ProcessManager:
        """
        Process manager used to run commands.
        """
        return self._process

    @property
    def terminal(self) -> TerminalManager:
        """
        Terminal manager used to create interactive terminals.
        """
        return self._terminal

    @property
    def filesystem(self) -> FilesystemManager:
        """
        Filesystem manager used to manage files.
        """
        return self._filesystem

    def __init__(
        self,
        template: str = "base",
        id: Optional[str] = None,
        api_key: Optional[str] = None,
        cwd: Optional[str] = None,
        env_vars: Optional[EnvVars] = None,
        on_scan_ports: Optional[Callable[[List[OpenPort]], Any]] = None,
        on_stdout: Optional[Callable[[ProcessMessage], Any]] = None,
        on_stderr: Optional[Callable[[ProcessMessage], Any]] = None,
        on_exit: Optional[Union[Callable[[int], Any], Callable[[], Any]]] = None,
        metadata: Optional[Dict[str, str]] = None,
        timeout: Optional[float] = TIMEOUT,
        _sandbox: Optional[models.Sandbox] = None,
        _debug_hostname: Optional[str] = None,
        _debug_port: Optional[int] = None,
        _debug_dev_env: Optional[Literal["remote", "local"]] = None,
    ):
        """
        Create a new cloud sandbox.

        :param id: [Deprecated] Use `template` param instead.
        :param template: ID of the sandbox template or the name of prepared template. If not specified a 'base' template will be used.
        Can be one of the following premade sandbox templates or a custom sandbox template ID:
        - `base` - A basic sandbox with a Linux environment
        - `Python3-DataAnalysis` - A Python3 sandbox with data analysis tools


        :param api_key: The API key to use, if not provided, the `E2B_API_KEY` environment variable is used
        :param cwd: The current working directory to use
        :param env_vars: A dictionary of environment variables to be used for all processes
        :param on_scan_ports: A callback to handle opened ports
        :param on_stdout: A default callback that is called when stdout with a newline is received from the process
        :param on_stderr: A default callback that is called when stderr with a newline is received from the process
        :param on_exit: A default callback that is called when the process exits
        :param metadata: A dictionary of strings that is stored alongside the running sandbox. You can see this metadata when you list running sandboxes.
        :param timeout: Timeout for sandbox to initialize in seconds, default is 60 seconds
        """

        template = id or template or "base"

        if id:
            logger.warning("The id parameter is deprecated, use template instead.")

        logger.info(
            f"Creating sandbox {template if isinstance(template, str) else type(template)}"
        )
        if cwd and cwd.startswith("~"):
            cwd = cwd.replace("~", "/home/user")

        self._code_snippet = CodeSnippetManager(
            sandbox=self,
            on_scan_ports=on_scan_ports,
        )

        self._on_stdout = on_stdout
        self._on_stderr = on_stderr

        default_env_vars = {"PYTHONUNBUFFERED": "1"}

        self._terminal = TerminalManager(sandbox=self)
        self._filesystem = FilesystemManager(sandbox=self)
        self._process = ProcessManager(
            sandbox=self,
            on_stdout=on_stdout,
            on_stderr=on_stderr,
            on_exit=on_exit,
        )
        super().__init__(
            template=template,
            api_key=api_key,
            cwd=cwd,
            env_vars={
                **default_env_vars,
                **(env_vars or {}),
            },
            metadata=metadata,
            _sandbox=_sandbox,
            _debug_hostname=_debug_hostname,
            _debug_port=_debug_port,
            _debug_dev_env=_debug_dev_env,
            timeout=timeout,
        )
        self._actions: Dict[str, Action[Self]] = {}

    def add_action(self, action: Action[Self], name: Optional[str] = None) -> "Sandbox":
        """
        Add a new action. If the name is not specified, it is automatically extracted from the function name.
        An action is a function that takes a sandbox and a dictionary of arguments and returns a string.

        You can use this action with specific integrations like OpenAI to interact with the sandbox and get output for the action.
        :param action: The action to add
        :param name: The name of the action, if not provided, the name of the function will be used

        Example:

            ```python
            from e2b import Sandbox

            def read_file(sandbox, args):
                with open(args["path"], "r") as f:
                    return sandbox.filesystem.read(args.path)

            s = Sandbox()
            s.add_action(read_file)
            s.add_action(name="hello", action=lambda s, args: f"Hello {args['name']}!")
            ```
        """

        if not name:
            name = action.__name__

        self._actions[name] = action

        return self

    def remove_action(self, name: str) -> "Sandbox":
        """
        Remove an action.

        :param name: The name of the action
        """
        del self._actions[name]

        return self

    @property
    def actions(self) -> Dict[str, Action[Self]]:
        """
        Return a dict of added actions.
        """

        return self._actions.copy()

    def action(self, name: Optional[str] = None):
        """
        Decorator to add an action.

        :param name: The name of the action, if not provided, the name of the function will be used
        """

        def _action(action: Action[Self]):
            self.add_action(action=action, name=name or action.__name__)

            return action

        return _action

    @property
    def openai(self):
        """
        OpenAI integration that can be used to get output for the actions added in the sandbox.

        Example:

            ```python
            from e2b import Sandbox

            s = Sandbox()
            s.openai.actions.run(run)
            ```
        """

        from e2b.templates.openai import OpenAI, Actions

        return OpenAI[Self](Actions[Self](self))

    def _handle_start_cmd_logs(self):
        def run_in_thread():
            self.process.start(
                "sudo journalctl --follow --lines=all -o cat _SYSTEMD_UNIT=start_cmd.service",
                cwd="/",
                env_vars={},
            )

        thread = threading.Thread(target=run_in_thread)
        thread.start()

    @classmethod
    def reconnect(
        cls,
        sandbox_id: str,
        cwd: Optional[str] = None,
        env_vars: Optional[EnvVars] = None,
        on_scan_ports: Optional[Callable[[List[OpenPort]], Any]] = None,
        on_stdout: Optional[Callable[[ProcessMessage], Any]] = None,
        on_stderr: Optional[Callable[[ProcessMessage], Any]] = None,
        on_exit: Optional[Union[Callable[[int], Any], Callable[[], Any]]] = None,
        timeout: Optional[float] = TIMEOUT,
        api_key: Optional[str] = None,
        _debug_hostname: Optional[str] = None,
        _debug_port: Optional[int] = None,
        _debug_dev_env: Optional[Literal["remote", "local"]] = None,
    ):
        """
        Reconnects to a previously created sandbox.

        :param sandbox_id: ID of the sandbox to reconnect to
        :param cwd: The current working directory to use
        :param env_vars: A dictionary of environment variables to be used for all processes
        :param on_scan_ports: A callback to handle opened ports
        :param on_stdout: A default callback that is called when stdout with a newline is received from the process
        :param on_stderr: A default callback that is called when stderr with a newline is received from the process
        :param on_exit: A default callback that is called when the process exits
        :param timeout: Timeout for sandbox to initialize in seconds, default is 60 seconds
        :param api_key: The API key to use, if not provided, the `E2B_API_KEY` environment variable is used


        ```py
        sandbox = Sandbox()
        id = sandbox.id
        sandbox.keep_alive(300)
        sandbox.close()

        # Reconnect to the sandbox
        reconnected_sandbox = Sandbox.reconnect(id)
        ```

        """

        logger.info(f"Reconnecting to sandbox {sandbox_id}")
        sandbox_id, client_id = sandbox_id.split("-")
        return cls(
            cwd=cwd,
            env_vars=env_vars,
            on_scan_ports=on_scan_ports,
            on_stdout=on_stdout,
            on_stderr=on_stderr,
            on_exit=on_exit,
            timeout=timeout,
            api_key=api_key,
            _sandbox=models.Sandbox(
                sandbox_id=sandbox_id,
                client_id=client_id,
                template_id=getattr(cls, "sandbox_template_id", "unknown"),
            ),
        )

    def _open(
        self,
        metadata: Optional[Dict[str, str]] = None,
        timeout: Optional[float] = TIMEOUT,
    ) -> None:
        """
        Open the sandbox.

        :param timeout: Specify the duration, in seconds to give the method to finish its execution before it times out (default is 60 seconds). If set to None, the method will continue to wait until it completes, regardless of time
        """
        logger.info(f"Opening sandbox {self._template}")
        super()._open(metadata=metadata, timeout=timeout)
        self._code_snippet._subscribe()
        logger.info(f"Sandbox {self._template} opened")

        if self.cwd:
            self.filesystem.make_dir(self.cwd)

        if self._on_stderr or self._on_stdout:
            self._handle_start_cmd_logs()

    def file_url(self) -> str:
        """
        Return a URL that can be used to upload files to the sandbox via a multipart/form-data POST request.
        This is useful if you're uploading files directly from the browser.
        The file will be uploaded to the user's home directory with the same name.
        If a file with the same name already exists, it will be overwritten.
        """
        hostname = self.get_hostname(self._debug_port or ENVD_PORT)
        protocol = self.get_protocol(secure=self._debug_dev_env != "local")

        file_url = f"{protocol}://{hostname}{FILE_ROUTE}"

        return file_url

    def upload_file(self, file: IO, timeout: Optional[float] = TIMEOUT) -> str:
        """
        Upload a file to the sandbox.
        The file will be uploaded to the user's home (`/home/user`) directory with the same name.
        If a file with the same name already exists, it will be overwritten.

        :param file: The file to upload
        :param timeout: Specify the duration, in seconds to give the method to finish its execution before it times out (default is 60 seconds). If set to None, the method will continue to wait until it completes, regardless of time
        """
        files = {"file": file}
        r = requests.post(self.file_url(), files=files, timeout=timeout)
        if r.status_code != 200:
            raise Exception(f"Failed to upload file: {r.reason} {r.text}")

        filename = path.basename(file.name)
        return f"/home/user/{filename}"

    def download_file(
        self, remote_path: str, timeout: Optional[float] = TIMEOUT
    ) -> bytes:
        """
        Download a file from the sandbox and returns it's content as bytes.

        :param remote_path: The path of the file to download
        :param timeout: Specify the duration, in seconds to give the method to finish its execution before it times out (default is 60 seconds). If set to None, the method will continue to wait until it completes, regardless of time
        """
        encoded_path = urllib.parse.quote(remote_path)
        url = f"{self.file_url()}?path={encoded_path}"
        r = requests.get(url, timeout=timeout)

        if r.status_code != 200:
            raise Exception(
                f"Failed to download file '{remote_path}'. {r.reason} {r.text}"
            )
        return r.content

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()



from typing import Any, Callable, ClassVar, List, Optional
from pydantic import BaseModel

from e2b.sandbox.exception import MultipleExceptions, RpcException, SandboxException
from e2b.sandbox.sandbox_connection import SandboxConnection, SubscriptionArgs


class OpenPort(BaseModel):
    ip: str
    port: int
    state: str


ScanOpenedPortsHandler = Callable[[List[OpenPort]], Any]


class CodeSnippetManager(BaseModel):
    # TODO: Change to model_config = ConfigDict(arbitrary_types_allowed=True) when only pydantic 2.0 is used
    class Config:
        arbitrary_types_allowed = True

    service_name: ClassVar[str] = "codeSnippet"
    sandbox: SandboxConnection

    on_scan_ports: Optional[Callable[[List[OpenPort]], Any]] = None

    def _subscribe(self):
        def on_scan_ports(ports: List[dict]):
            ports = [
                OpenPort(
                    ip=port["Ip"],
                    port=port["Port"],
                    state=port["State"],
                )
                for port in ports
            ]
            if self.on_scan_ports:
                self.on_scan_ports(ports)

        if self.on_scan_ports:
            try:
                self.sandbox._handle_subscriptions(
                    SubscriptionArgs(
                        service=self.service_name,
                        handler=on_scan_ports,
                        method="scanOpenedPorts",
                    )
                )
            except RpcException as e:
                raise SandboxException(e.message) from e
            except MultipleExceptions as e:
                raise SandboxException("Failed to subscribe to RPC services") from e

        return self



"""
Secure sandboxed cloud environments made for AI agents and AI apps

Check usage docs - https://e2b.dev/docs/sandbox/overview

E2B Sandbox is a secure sandboxed cloud environment made for AI agents and AI
apps. Sandboxes allow AI agents and apps to have long running cloud secure
environments. In these environments, large language models can use the same
tools as humans do.


```py
from e2b import Sandbox

# Create sandbox
sandbox = Sandbox()

# Let an LLM use the sandbox here

# Close sandbox once done
sandbox.close()
```
"""

from .api import (
    E2BApiClient,
    client,
)
from .constants import (
    SANDBOX_DOMAIN,
    API_HOST,
)
from .sandbox import (
    Sandbox,
    FilesystemOperation,
    FilesystemWatcher,
    FileInfo,
    FilesystemEvent,
    FilesystemManager,
    TerminalManager,
    Terminal,
    ProcessManager,
    Process,
    OpenPort,
    Action,
    EnvVars,
    SandboxException,
    TerminalException,
    ProcessException,
    CurrentWorkingDirectoryDoesntExistException,
    FilesystemException,
    RpcException,
    ProcessMessage,
    ProcessOutput,
    TerminalOutput,
    run_code,
    RunningSandbox,
)
from .templates import (
    DataAnalysis,
    CodeInterpreter,
)



# coding: utf-8

# flake8: noqa

"""
    E2B API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: 0.1.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


__version__ = "1.0.0"

# import apis into sdk package
from e2b.api.v2.client.api.sandboxes_api import SandboxesApi

# import ApiClient
from e2b.api.v2.client.api_response import ApiResponse
from e2b.api.v2.client.api_client import ApiClient
from e2b.api.v2.client.configuration import Configuration
from e2b.api.v2.client.exceptions import OpenApiException
from e2b.api.v2.client.exceptions import ApiTypeError
from e2b.api.v2.client.exceptions import ApiValueError
from e2b.api.v2.client.exceptions import ApiKeyError
from e2b.api.v2.client.exceptions import ApiAttributeError
from e2b.api.v2.client.exceptions import ApiException

# import models into sdk package
from e2b.api.v2.client.models.environment import Environment
from e2b.api.v2.client.models.environment_build import EnvironmentBuild
from e2b.api.v2.client.models.error import Error
from e2b.api.v2.client.models.instance import Instance
from e2b.api.v2.client.models.instances_instance_id_refreshes_post_request import (
    InstancesInstanceIDRefreshesPostRequest,
)
from e2b.api.v2.client.models.new_instance import NewInstance
from e2b.api.v2.client.models.new_sandbox import NewSandbox
from e2b.api.v2.client.models.running_instance import RunningInstance
from e2b.api.v2.client.models.running_sandboxes import RunningSandboxes
from e2b.api.v2.client.models.sandbox import Sandbox
from e2b.api.v2.client.models.sandboxes_sandbox_id_refreshes_post_request import (
    SandboxesSandboxIDRefreshesPostRequest,
)
from e2b.api.v2.client.models.template import Template
from e2b.api.v2.client.models.template_build import TemplateBuild
from e2b.api.v2.client.models.templates_template_id_builds_build_id_logs_post_request import (
    TemplatesTemplateIDBuildsBuildIDLogsPostRequest,
)



import logging

from enum import Enum
from typing import Any, Callable, Optional, Set

from pydantic import BaseModel

from e2b.constants import TIMEOUT
from e2b.sandbox.exception import FilesystemException, RpcException
from e2b.sandbox.sandbox_connection import SandboxConnection
from e2b.utils.str import snake_case_to_camel_case

logger = logging.getLogger(__name__)


class FilesystemOperation(str, Enum):
    Create = "Create"
    Write = "Write"
    Remove = "Remove"
    Rename = "Rename"
    Chmod = "Chmod"


class FilesystemEvent(BaseModel):
    path: str
    name: str
    operation: FilesystemOperation
    timestamp: int
    """
    Unix epoch in nanoseconds
    """
    is_dir: bool

    class Config:
        alias_generator = snake_case_to_camel_case


class FilesystemWatcher:
    @property
    def path(self) -> str:
        """
        The path to watch.
        """
        return self._path

    def __init__(
        self,
        connection: SandboxConnection,
        path: str,
        service_name: str,
    ):
        self._connection = connection
        self._path = path
        self._service_name = service_name
        self._unsubscribe: Optional[Callable[[], Any]] = None
        self._listeners: Set[Callable[[FilesystemEvent], Any]] = set()

    def start(self, timeout: Optional[float] = TIMEOUT) -> None:
        """
        Start the filesystem watcher.

        :param timeout: Specify the duration, in seconds to give the method to finish its execution before it times out (default is 60 seconds). If set to None, the method will continue to wait until it completes, regardless of time
        """
        if self._unsubscribe:
            return

        logger.debug("Starting filesystem watcher for %s", self.path)
        try:
            self._unsubscribe = self._connection._subscribe(
                self._service_name,
                self._handle_filesystem_events,
                "watchDir",
                self.path,
                timeout=timeout,
            )
            logger.debug("Started filesystem watcher for %s", self.path)
        except RpcException as e:
            raise FilesystemException(e.message) from e

    def stop(self) -> None:
        """
        Stop the filesystem watcher.
        """
        logger.debug("Stopping filesystem watcher for %s", self.path)

        self._listeners.clear()
        if self._unsubscribe:
            try:
                self._unsubscribe()
                self._unsubscribe = None
                logger.debug("Stopped filesystem watcher for %s", self.path)
            except RpcException as e:
                raise FilesystemException(e.message) from e

    def add_event_listener(self, listener: Callable[[FilesystemEvent], Any]):
        """
        Add a listener for filesystem events.

        :param listener: Listener to add

        :return: Function that removes the listener
        """
        logger.debug("Adding filesystem watcher listener for %s", self.path)

        self._listeners.add(listener)

        def delete_listener() -> None:
            self._listeners.remove(listener)

        return delete_listener

    def _handle_filesystem_events(self, event: dict) -> None:
        event = FilesystemEvent(**event)
        for listener in self._listeners:
            listener(event)



import concurrent
import logging
import threading
import traceback
import warnings

from concurrent.futures import ThreadPoolExecutor, CancelledError
from time import sleep
from typing import Any, Callable, List, Literal, Optional, Union, Dict
from datetime import datetime
from pydantic import BaseModel
from urllib3.exceptions import ReadTimeoutError, MaxRetryError, ConnectTimeoutError

from e2b.api import E2BApiClient, exceptions, models, client
from e2b.constants import (
    SANDBOX_DOMAIN,
    SANDBOX_REFRESH_PERIOD,
    TIMEOUT,
    ENVD_PORT,
    WS_ROUTE,
    SECURE,
)
from e2b.sandbox.env_vars import EnvVars
from e2b.sandbox.exception import (
    MultipleExceptions,
    SandboxException,
    TimeoutException,
    SandboxNotOpenException,
)
from e2b.sandbox.sandbox_rpc import Notification, SandboxRpc
from e2b.utils.api_key import get_api_key
from e2b.utils.future import DeferredFuture
from e2b.utils.str import camel_case_to_snake_case

logger = logging.getLogger(__name__)


class Subscription(BaseModel):
    service: str
    id: str
    handler: Callable[[Any], None]


class SubscriptionArgs(BaseModel):
    service: str
    handler: Callable[[Any], None]
    method: str
    params: List[Any] = []


class RunningSandbox(BaseModel):
    sandbox_id: str
    template_id: str
    alias: Optional[str]
    metadata: Optional[Dict[str, str]]
    cpu_count: int
    memory_mb: int
    started_at: datetime


class SandboxConnection:
    _refresh_retries = 4
    _on_close_child: Optional[Callable[[], Any]] = None

    @property
    def id(self) -> str:
        """
        The sandbox ID.

        You can use this ID to reconnect to the sandbox later.
        """
        if not self._sandbox:
            raise SandboxException("Sandbox is not running.")
        return f"{self._sandbox.sandbox_id}-{self._sandbox.client_id}"

    @property
    def finished(self):
        """
        A future that is resolved when the sandbox exits.
        """
        return self._finished

    @property
    def is_open(self) -> bool:
        """
        Whether the sandbox is open.
        """
        return self._is_open

    def __init__(
        self,
        template: str,
        api_key: Optional[str] = None,
        cwd: Optional[str] = None,
        env_vars: Optional[EnvVars] = None,
        metadata: Optional[Dict[str, str]] = None,
        timeout: Optional[float] = TIMEOUT,
        _sandbox: Optional[models.Sandbox] = None,
        _debug_hostname: Optional[str] = None,
        _debug_port: Optional[int] = None,
        _debug_dev_env: Optional[Literal["remote", "local"]] = None,
    ):
        api_key = get_api_key(api_key)

        self.cwd = cwd
        """
        Default working directory used in the sandbox.

        You can change the working directory by setting the `cwd` property.
        """
        self.env_vars = env_vars or {}
        """
        Default environment variables used in the sandbox.

        You can change the environment variables by setting the `env_vars` property.
        """
        self._template = template
        self._api_key = api_key
        self._debug_hostname = _debug_hostname
        self._debug_port = _debug_port
        self._debug_dev_env = _debug_dev_env
        self._sandbox = _sandbox

        self._is_open = False
        self._process_cleanup: List[Callable[[], Any]] = []
        self._subscribers = {}
        self._rpc: Optional[SandboxRpc] = None
        self._finished = DeferredFuture(self._process_cleanup)

        logger.info(f"Sandbox for template {self._template} initialized")

        self._open(metadata=metadata, timeout=timeout)

    @classmethod
    def create(cls, *args, **kwargs):
        warnings.warn("Sandbox.create() is deprecated, use Sandbox() instead")
        return cls(*args, **kwargs)

    def get_hostname(self, port: Optional[int] = None) -> str:
        """
        Get the hostname for the sandbox or for the specified sandbox's port.

        :param port: Specify if you want to connect to a specific port of the sandbox

        :return: Hostname of the sandbox or sandbox's port
        """

        if self._debug_hostname:
            if port and self._debug_dev_env == "remote":
                return f"{port}-{self._debug_hostname}"
            elif port:
                return f"{self._debug_hostname}:{port}"
            else:
                return self._debug_hostname
        else:
            if not self._sandbox:
                raise SandboxException("Sandbox is not running.")

        hostname = f"{self.id}.{SANDBOX_DOMAIN}"

        if port:
            return f"{port}-{hostname}"

        return hostname

    @staticmethod
    def get_protocol(base_protocol: str = "http", secure: bool = SECURE) -> str:
        """
        The function decides whether to use the secure or insecure protocol.

        :param base_protocol: Specify the specific protocol you want to use. Do not include the `s` in `https` or `wss`.
        :param secure: Specify whether you want to use the secure protocol or not.

        :return: Protocol for the connection to the sandbox
        """
        return f"{base_protocol}s" if secure else base_protocol

    def keep_alive(self, duration: int) -> None:
        """
        Keep the sandbox alive for the specified duration.

        :param duration: Duration in seconds. Must be between 0 and 3600 seconds.
        """
        if not 0 <= duration <= 3600:
            raise ValueError("Duration must be between 0 and 3600 seconds")

        with E2BApiClient(api_key=self._api_key) as api_client:
            api = client.SandboxesApi(api_client)
            try:
                api.sandboxes_sandbox_id_refreshes_post(
                    self._sandbox.sandbox_id,
                    client.SandboxesSandboxIDRefreshesPostRequest(duration=duration),
                )
                logger.debug(
                    f"Sandbox will be kept alive without connection for next {duration} seconds."
                )
            except exceptions.ApiException as e:
                if e.status == 404:
                    raise SandboxException(
                        f"Sandbox {self.id} failed because it cannot be found"
                    ) from e
                else:
                    raise e

    def close(self) -> None:
        """
        Close the sandbox and unsubscribe from all the subscriptions.
        """
        self._close()
        logger.info(f"Sandbox closed")

    def _close(self):
        if self._is_open and self._sandbox:
            logger.info(f"Closing sandbox {self._sandbox.template_id} (id: {self.id})")
            self._is_open = False
            if self._rpc:
                self._rpc.close()

        if self._on_close_child:
            self._on_close_child()

        for cleanup in self._process_cleanup:
            cleanup()
        self._process_cleanup.clear()

    def _open(
        self,
        metadata: Optional[Dict[str, str]] = None,
        timeout: Optional[float] = TIMEOUT,
    ) -> None:
        """
        Open a connection to a new sandbox.

        You must call this method before using the sandbox.
        """
        if self._is_open:
            raise SandboxException("Sandbox connect was already called")
        else:
            self._is_open = True

        if not self._sandbox and not self._debug_hostname:
            try:
                with E2BApiClient(api_key=self._api_key) as api_client:
                    api = client.SandboxesApi(api_client)

                    self._sandbox = api.sandboxes_post(
                        models.NewSandbox(templateID=self._template, metadata=metadata),
                        _request_timeout=timeout,
                    )
                    logger.info(
                        f"Sandbox {self._sandbox.template_id} created (id:{self.id})"
                    )
            except ReadTimeoutError as e:
                logger.error(f"Failed to acquire sandbox")
                self._close()
                raise TimeoutException(
                    f"Failed to acquire sandbox: {e}",
                ) from e
            except MaxRetryError as e:
                if isinstance(e.reason, ConnectTimeoutError):
                    raise TimeoutException(
                        f"Failed to acquire sandbox: {e}",
                    ) from e
                raise e
            except Exception as e:
                logger.error(f"Failed to acquire sandbox")
                self._close()
                raise e

        if not self._debug_hostname:
            self._start_refreshing()

        try:
            self._connect_rpc(timeout)
        except Exception as e:
            logger.error(e)
            self._close()
            raise e

    def _connect_rpc(self, timeout: Optional[float] = TIMEOUT):
        hostname = self.get_hostname(self._debug_port or ENVD_PORT)
        protocol = self.get_protocol("ws", self._debug_dev_env != "local")
        sandbox_url = f"{protocol}://{hostname}{WS_ROUTE}"

        self._rpc = SandboxRpc(
            url=sandbox_url,
            on_message=self._handle_notification,
        )
        self._rpc.connect(timeout=timeout)

    def _start_refreshing(self):
        threading.Thread(
            target=self._refresh,
            daemon=True,
            name="e2b-sandbox-refresh",
        ).start()

    def _call(
        self,
        service: str,
        method: str,
        params: Optional[List[Any]] = None,
        timeout: Optional[float] = TIMEOUT,
    ) -> Any:
        if not params:
            params = []

        if not self.is_open:
            raise SandboxNotOpenException("Sandbox is not open")

        if not self._rpc:
            raise SandboxException("Sandbox is not connected")

        return self._rpc.send_message(f"{service}_{method}", params, timeout)

    def _handle_subscriptions(
        self,
        *subscription_args: SubscriptionArgs,
    ):
        results: List[Union[Callable, None, Exception]] = []
        with ThreadPoolExecutor(thread_name_prefix="process-subscribe") as executor:
            futures = []
            for args in subscription_args:
                future = executor.submit(
                    self._subscribe,
                    args.service,
                    args.handler,
                    args.method,
                    *args.params,
                )
                futures.append(future)

            for future in concurrent.futures.as_completed(futures):
                results.append(future.result())

        results = [sub for sub in results if sub]
        process_exceptions = [e for e in results if isinstance(e, Exception)]

        def unsub_all():
            def unsub_func():
                for unsub in results:
                    if not isinstance(unsub, Exception):
                        try:
                            unsub()
                        except (CancelledError, SandboxNotOpenException, KeyError):
                            pass

            threading.Thread(
                name="process-unsubscribe",
                daemon=True,
                target=unsub_func,
            ).start()

        if len(process_exceptions) > 0:
            unsub_all()

            if len(process_exceptions) == 1:
                raise process_exceptions[0]

            error_message = "\n"

            for i, s in enumerate(process_exceptions):
                tb = s.__traceback__  # Get the traceback object
                stack_trace = "\n".join(traceback.extract_tb(tb).format())
                error_message += f'\n[{i}]: {type(s).__name__}("{s}"):\n{stack_trace}\n'

            raise MultipleExceptions(
                message=error_message,
                exceptions=process_exceptions,
            )

        return unsub_all

    def _unsubscribe(self, sub_id: str, timeout: Optional[float] = TIMEOUT):
        sub = self._subscribers[sub_id]
        self._call(sub.service, "unsubscribe", [sub.id], timeout=timeout)
        del self._subscribers[sub_id]
        logger.debug(f"Unsubscribed (sub_id: {sub_id})")

    def _subscribe(
        self,
        service: str,
        handler: Callable[[Any], Any],
        method: str,
        *params,
        timeout: Optional[float] = TIMEOUT,
    ) -> Callable[[], None]:
        sub_id = self._call(service, "subscribe", [method, *params], timeout=timeout)
        if not isinstance(sub_id, str):
            raise Exception(f"Failed to subscribe: {camel_case_to_snake_case(method)}")

        self._subscribers[sub_id] = Subscription(
            service=service, id=sub_id, handler=handler
        )
        logger.debug(
            f"Subscribed to {service} {camel_case_to_snake_case(method)} (sub id: {sub_id})"
        )

        def unsub():
            self._unsubscribe(sub_id, timeout=timeout)

        return unsub

    def _handle_notification(self, data: Notification):
        logger.debug(f"Notification {data}")

        for id, sub in self._subscribers.items():
            if id == data.params["subscription"]:
                sub.handler(data.params["result"])

    def _refresh(self):
        if not self._sandbox:
            logger.info("No sandbox to refresh. Sandbox was not created")
            return

        try:
            logger.info(
                f"Started refreshing sandbox {self._sandbox.template_id} (id: {self.id})"
            )

            current_retry = 0

            with E2BApiClient(api_key=self._api_key) as api_client:
                api = client.SandboxesApi(api_client)
                while True:
                    if not self._is_open:
                        logger.debug(
                            f"Cannot refresh sandbox - it was closed. {self.id}"
                        )
                        return
                    sleep(SANDBOX_REFRESH_PERIOD)
                    try:
                        api.sandboxes_sandbox_id_refreshes_post(
                            self._sandbox.sandbox_id,
                            client.SandboxesSandboxIDRefreshesPostRequest(duration=0),
                        )
                        logger.debug(f"Refreshed sandbox {self.id}")
                    except exceptions.ApiException as e:
                        if e.status == 404:
                            raise SandboxException(
                                f"Sandbox {self.id} failed because it cannot be found"
                            ) from e
                        else:
                            if current_retry < self._refresh_retries:
                                logger.error(
                                    f"Refreshing sandbox {self.id} failed. Retrying..."
                                )
                                current_retry += 1
                            else:
                                logger.error(
                                    f"Refreshing sandbox {self.id} failed. Max retries exceeded"
                                )
                                raise e
        finally:
            if self._sandbox:
                logger.info(f"Stopped refreshing sandbox (id: {self.id})")
            else:
                logger.info("No sandbox to stop refreshing. Sandbox was not created")

            self._close()

    @staticmethod
    def list(api_key: Optional[str] = None) -> List[RunningSandbox]:
        """
        List all running sandboxes.

        :param api_key: API key to use for authentication.
        If not provided, the `E2B_API_KEY` environment variable will be used.
        """
        api_key = get_api_key(api_key)

        with E2BApiClient(api_key=api_key) as api_client:
            return [
                RunningSandbox(
                    sandbox_id=f"{sandbox.sandbox_id}-{sandbox.client_id}",
                    **sandbox.dict(exclude={"sandbox_id", "client_id"}),
                )
                for sandbox in client.SandboxesApi(api_client).sandboxes_get()
            ]

    @staticmethod
    def kill(sandbox_id: str, api_key: Optional[str] = None) -> None:
        """
        Kill the running sandbox specified by the sandbox ID.

        :param sandbox_id: ID of the sandbox to kill.
        :param api_key: API key to use for authentication.
        If not provided, the `E2B_API_KEY` environment variable will be used.
        """
        api_key = get_api_key(api_key)

        short_id = sandbox_id.split("-")[0]
        with E2BApiClient(api_key=api_key) as api_client:
            return client.SandboxesApi(api_client).sandboxes_sandbox_id_delete(
                sandbox_id=short_id
            )



import os

SANDBOX_REFRESH_PERIOD = 5  # seconds

TIMEOUT = 60

DOMAIN = os.getenv("E2B_DOMAIN") or "e2b.dev"
SECURE = os.getenv("E2B_SECURE", "TRUE").upper() == "TRUE"
DEBUG = os.getenv("E2B_DEBUG") or False

API_DOMAIN = "localhost:3000" if DEBUG else f"api.{DOMAIN}"
PROTOCOL = "https" if SECURE and not DEBUG else "http"
API_HOST = f"{PROTOCOL}://{API_DOMAIN}"


SANDBOX_DOMAIN = DOMAIN

ENVD_PORT = 49982
WS_ROUTE = "/ws"
FILE_ROUTE = "/file"



# coding: utf-8

# flake8: noqa
"""
    E2B API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: 0.1.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


# import models into model package
from e2b.api.v1.client.models.environment import Environment
from e2b.api.v1.client.models.environment_build import EnvironmentBuild
from e2b.api.v1.client.models.error import Error
from e2b.api.v1.client.models.instance import Instance
from e2b.api.v1.client.models.instances_instance_id_refreshes_post_request import (
    InstancesInstanceIDRefreshesPostRequest,
)
from e2b.api.v1.client.models.new_instance import NewInstance
from e2b.api.v1.client.models.new_sandbox import NewSandbox
from e2b.api.v1.client.models.running_instance import RunningInstance
from e2b.api.v1.client.models.running_sandboxes import RunningSandboxes
from e2b.api.v1.client.models.sandbox import Sandbox
from e2b.api.v1.client.models.sandboxes_sandbox_id_refreshes_post_request import (
    SandboxesSandboxIDRefreshesPostRequest,
)
from e2b.api.v1.client.models.template import Template
from e2b.api.v1.client.models.template_build import TemplateBuild
from e2b.api.v1.client.models.templates_template_id_builds_build_id_logs_post_request import (
    TemplatesTemplateIDBuildsBuildIDLogsPostRequest,
)



import queue
import sys

from concurrent.futures import ThreadPoolExecutor


def shutdown_executor(executor: ThreadPoolExecutor):
    py_version = sys.version_info
    if (py_version.major == 3) and (py_version.minor < 9):
        # py versions less than 3.9
        # Executor#shutdown does not accept
        # cancel_futures keyword
        # manually shutdown
        # code taken from https://github.com/python/cpython/blob/3.9/Lib/concurrent/futures/thread.py#L216

        # prevent new tasks from being submitted
        executor.shutdown(wait=False)
        while True:
            # cancel all waiting tasks
            try:
                work_item = executor._work_queue.get_nowait()

            except queue.Empty:
                break

            if work_item is not None:
                work_item.future.cancel()

    else:
        executor.shutdown(wait=False, cancel_futures=True)
